"""
Suite2p Result Parser & Utils
===============================

:author:
    Yu-Ting Wei

This module provide the container for suite2p registration/segmentation 2P results.

Also include utilities for signal processing and basic plotting

"""
from .core import *
from .plot import *
from .signals import *
